from typing import Any

from .exceptions import SchemaError


def parse_values_map(data: list) -> dict[str, Any]:
    """Parse a values map object into a dictionary.

    The values map object is a list of objects, each with a 'type' and 'values' key.
    Technically, the 'key' can be duplicated, but we assume that it is unique. If we find
    documentation that says otherwise, we can change this. But it would be nice to know if
    the values of duplicate keys can be combined or if they must be handled separately before
    we make that change.

    Parameters
    ----------
    data : list
        The values map object from a JSON.

    Returns
    -------
    dict
        The parsed values map object as a dictionary.
    """
    if not isinstance(data, list):
        raise SchemaError("Expected 'valuesMap' to be an array.")

    result = {}
    for item in data:
        if not isinstance(item, dict):
            raise SchemaError("Expected 'valuesMap' to contain objects.")

        if "type" not in item:
            raise SchemaError("Missing 'type' in values map schema.")

        if "values" not in item:
            raise SchemaError(f"Missing 'values' in values map schema for '{item['type']}'.")

        if not isinstance(item["values"], list):
            raise SchemaError(f"Expected 'values' for '{item['type']}' to be an array.")

        if item["type"] in result:
            raise SchemaError(f"Duplicate type '{item['type']}' in values map schema.")

        result[item["type"]] = []
        for value in item["values"]:
            match value:
                case bool() | float() | int() | str():
                    result[item["type"]].append(value)
                case {"x": x, "y": y}:  # Point
                    # we currently store this as a tuple
                    result[item["type"]].append((x, y))
                case _:
                    msg = f"Unsupported type for '{item['type']}': '{value}' in values map schema"
                    raise SchemaError(msg)
    return result



import datetime

from .exceptions import SchemaError
from .interval import Interval
from .interval_period import IntervalPeriod
from .payload_descriptor import PayloadDescriptor
from .report_descriptor import ReportDescriptor
from .values_map import parse_values_map


class Event:
    """Event object to communicate a Demand Response request to VEN.

    If intervalPeriod is present, sets start time and duration of intervals.
    """

    def __init__(self, data: dict):
        if "objectType" in data and data["objectType"] != "EVENT":
            raise SchemaError("Expected 'objectType' to be 'EVENT' in event schema.")

        if "programID" not in data:
            raise SchemaError("Missing 'programID' in event schema.")

        if "intervals" not in data:
            raise SchemaError("Missing 'intervals' in event schema.")

        self._id = data.get("id", None)
        self._program_id = data["programID"]
        self._intervals = [Interval(interval) for interval in data["intervals"]]
        self._name = data.get("eventName", None)
        self._interval_period = None
        self._created = None
        self._modified = None
        self._priority = None
        self._targets = None
        self._report_descriptors = None
        self._payload_descriptors = None

        if "intervalPeriod" in data:
            self._interval_period = IntervalPeriod(data["intervalPeriod"])

        if "createdDateTime" in data:
            self._created = datetime.datetime.fromisoformat(data["createdDateTime"])

        if "modificationDateTime" in data:
            self._modified = datetime.datetime.fromisoformat(data["modificationDateTime"])

        if "priority" in data:
            try:
                self._priority = int(data["priority"])
            except ValueError as err:
                msg = "Expected 'priority' to be an integer in event schema."
                raise SchemaError(msg) from err

        if "targets" in data:
            self._targets = parse_values_map(data["targets"])

        if "reportDescriptors" in data:
            self._report_descriptors = [
                ReportDescriptor(report) for report in data["reportDescriptors"]
            ]

        if "payloadDescriptors" in data:
            self._payload_descriptors = [
                PayloadDescriptor(payload) for payload in data["payloadDescriptors"]
            ]

    @property
    def id(self) -> str | None:
        """The ID of the event."""
        return self._id

    @property
    def event_name(self) -> str | None:
        """The name of the event."""
        return self._name

    @property
    def created(self) -> datetime.datetime | None:
        """The time the event was created."""
        return self._created

    @property
    def modified(self) -> datetime.datetime | None:
        """The time the event was last modified."""
        return self._modified

    @property
    def interval_period(self) -> IntervalPeriod | None:
        """Defines default start and durations of intervals."""
        return self._interval_period

    @property
    def intervals(self) -> list[Interval]:
        """List of interval objects."""
        return self._intervals

    @property
    def program_id(self) -> str:
        """The ID of the program object this event is associated with."""
        return self._program_id

    @property
    def priority(self) -> int | None:
        """The priority of the event."""
        return self._priority

    @property
    def targets(self) -> dict[str, list] | None:
        """The targets of the event."""
        return self._targets

    @property
    def report_descriptors(self) -> list[ReportDescriptor] | None:
        """The report descriptors of the event."""
        return self._report_descriptors

    @property
    def payload_descriptors(self) -> list[PayloadDescriptor] | None:
        """The payload descriptors of the event."""
        return self._payload_descriptors



from enum import Enum

import aiohttp

from .access_token import AccessToken
from .event import Event
from .exceptions import ToadrError


class TargetType(Enum):
    """Enumeration of target types."""

    POWER_SERVICE_LOCATION = "POWER_SERVICE_LOCATION"
    """
    A Power Service Location is a utility named specific location in geography or
    the distribution system, usually the point of service to a customer site.
    """
    SERVICE_AREA = "SERVICE_AREA"
    """
    A Service Area is a utility named geographic region.
    Target values array contains a string representing a service area name.
    """
    GROUP = "GROUP"
    """
    Target values array contains a string representing a group.
    """
    RESOURCE_NAME = "RESOURCE_NAME"
    """
    Target values array contains a string representing a resource name.
    """
    VEN_NAME = "VEN_NAME"
    """
    Target values array contains a string representing a VEN name.
    """
    EVENT_NAME = "EVENT_NAME"
    """
    Target values array contains a string representing an event name.
    """
    PROGRAM_NAME = "PROGRAM_NAME"
    """
    Target values array contains a string representing a program name.
    """


async def get_events(
    session: aiohttp.ClientSession,
    vtn_url: str,
    access_token: AccessToken | None,
    program_id: str | None = None,
    target_type: TargetType | None = None,
    target_values: list[str] | None = None,
    skip: int | None = None,
    limit: int | None = None,
) -> list[Event]:
    """Get a list of events from the VTN.

    Lists all the events available from the VTN.
    The results can be filtered by providing a program ID, and/or target type and target values.

    Parameters
    ----------
    session : aiohttp.ClientSession
        The aiohttp session to use for the request.
    vtn_url : str
        The URL of the VTN.
    access_token : AccessToken | None
        The access token to use for the request, use None if no token is required.
    program_id : int | None
        The program ID to filter the events by.
    target_type : TargetType | None
        The target type to filter the events by.
    target_values : list[str] | None
        The target values to filter the events by (names of the target type).
    skip : int | None
        The number of events to skip (for pagination).
    limit : int | None
        The maximum number of events to return.

    Returns
    -------
    list[Event]
        A list of events.

    Raises
    ------
    ValueError
        If the query parameters are invalid.
    toadr3.ToadrError
        If the request to the VTN fails.
    """
    _check_arguments(target_type, target_values, skip, limit)

    params = _create_query_parameters(program_id, target_type, target_values, skip, limit)

    headers = {}
    if access_token is not None:
        headers["Authorization"] = f"Bearer {access_token.token}"

    while vtn_url.endswith("/"):
        vtn_url = vtn_url[:-1]

    async with session.get(f"{vtn_url}/events", params=params, headers=headers) as response:
        # 400 is start of HTTP error codes
        if response.status >= 400:  # noqa PLR2004 - Magic value used in comparison
            match response.status:
                case 400 | 403 | 500:
                    json_response = await response.json()  # JSON should be of type Problem schema
                    message = json_response["title"]

                    if "detail" in json_response:
                        message = f"{message} - {json_response['detail']}"

                    raise ToadrError(
                        message,
                        status_code=response.status,
                        reason=response.reason,
                        headers=response.headers,
                        json_response=json_response,
                    )
                case _:
                    raise RuntimeError(
                        f"Unexpected response status: {response.status} {response.reason}"
                    )

        data = await response.json()

        result = []
        for event in data:
            result.append(Event(event))
        return result


def _create_query_parameters(
    program_id: int | None,
    target_type: TargetType | None,
    target_values: list[str] | None,
    skip: int | None,
    limit: int | None,
) -> dict[str, int | str | list[str]]:
    """Create the query parameters for the get_events function.

    Parameters
    ----------
    program_id : int | None
        The program ID to filter the events by.
    target_type : TargetType | None
        The target type to filter the events by.
    target_values : list[str] | None
        The target values to filter the events by (names of the target type).
    skip : int | None
        The number of events to skip (for pagination).
    limit : int | None
        The maximum number of events to return.
    """
    params = {}
    if program_id is not None:
        params["programID"] = program_id

    if target_type is not None:
        params["targetType"] = target_type.value
        params["targetValues"] = target_values

    if skip is not None:
        params["skip"] = skip

    if limit is not None:
        params["limit"] = limit

    return params


def _check_arguments(
    target_type: TargetType | None,
    target_values: list[str] | None,
    skip: int | None,
    limit: int | None,
):
    """
    Check the arguments for the get_events function.

    Parameters
    ----------
    target_type : TargetType | None
        The target type to filter the events by.
    target_values : list[str] | None
        The target values to filter the events by (names of the target type).
    skip : int | None
        The number of events to skip (for pagination).
    limit : int | None
        The maximum number of events to return.

    Raises
    ------
    ValueError
        If the query parameters are invalid.
    """
    if target_type is not None and target_values is None:
        raise ValueError("target_values are required when target_type is provided")

    if target_values is not None and target_type is None:
        raise ValueError("target_type is required when target_values are provided")

    if target_values is not None and not isinstance(target_values, list):
        raise ValueError("target_values must be a list of strings")

    if skip is not None and not isinstance(skip, int):
        raise ValueError("skip must be an integer")

    if skip is not None and skip < 0:
        raise ValueError("skip must be a positive integer")

    if limit is not None and not isinstance(limit, int):
        raise ValueError("limit must be an integer")

    if limit is not None and not limit >= 0:
        raise ValueError("limit must be a positive integer")



from .exceptions import SchemaError


class PayloadDescriptor:
    """Payload descriptor object to describe the payload of an Event."""

    def __init__(self, data: dict):
        """Create a payload descriptor object from parsed JSON data."""
        if "payloadType" not in data:
            raise SchemaError("Missing 'payloadType' in payload descriptor schema.")

        self._payload_type = data["payloadType"]
        self._units: str | None = data.get("units", None)
        self._currency: str | None = data.get("currency", None)

    @property
    def payload_type(self) -> str:
        """The type of the payload."""
        return self._payload_type

    @property
    def units(self) -> str | None:
        """The measurement units."""
        return self._units

    @property
    def currency(self) -> str | None:
        """Currency of price payload."""
        return self._currency



class ToadrError(Exception):
    """Exception raised for errors in the Toadr API."""

    def __init__(
        self,
        message: str,
        status_code: int | str,
        reason: str | None = None,
        headers: dict | None = None,
        json_response: dict | None = None,
    ):
        self._message = message

        if isinstance(status_code, str):
            status_code = int(status_code)

        self._status_code: int = status_code
        self._reason = reason
        self._headers = headers
        self._json_response = json_response

    @property
    def message(self) -> str:
        """A short description of the exception."""
        return self._message

    @property
    def status_code(self) -> int:
        """The HTTP status code of the response."""
        return self._status_code

    @property
    def reason(self) -> str:
        """The reason of the exception (message from aiohttp.ClientResponseError)."""
        return self._reason

    @property
    def headers(self) -> dict:
        """The headers of the response."""
        return self._headers

    @property
    def json_response(self) -> dict:
        """The JSON response from the request (full JSON error response, if available)."""
        return self._json_response


class SchemaError(Exception):
    """Exception raised for schema errors in the Toadr API."""

    def __init__(self, message: str):
        self._message = message

    @property
    def message(self) -> str:
        """A short description of the exception."""
        return self._message



import datetime

from .exceptions import SchemaError
from .interval import Interval
from .interval_period import IntervalPeriod
from .payload_descriptor import PayloadDescriptor


class ReportData:
    """Report data for a resource."""

    def __init__(self, data: dict):
        """Create a new report data object from parsed JSON data.."""
        if "resourceName" not in data:
            raise SchemaError("Missing 'resourceName' in report data schema.")

        if "intervals" not in data:
            raise SchemaError("Missing 'intervals' in report data schema.")

        self._resource_name = data["resourceName"]
        self._intervals = [Interval(interval) for interval in data["intervals"]]
        self._interval_period = None

        if "intervalPeriod" in data:
            self._interval_period = IntervalPeriod(data["intervalPeriod"])

    @property
    def resource_name(self) -> str:
        """The name of the resource."""
        return self._resource_name

    @property
    def intervals(self) -> list[Interval]:
        """Applicable intervals for this report and resource."""
        return self._intervals

    @property
    def interval_period(self) -> IntervalPeriod | None:
        """The interval period for this report and resource (if available)."""
        return self._interval_period


class Report:
    """Report object for a report."""

    def __init__(self, data: dict):
        """Create a new report object from parsed JSON data."""
        if "objectType" in data and data["objectType"] != "REPORT":
            raise SchemaError("Expected 'objectType' to be 'REPORT' in report data schema.")

        if "programID" not in data:
            raise SchemaError("Missing 'programID' in report schema.")

        if "eventID" not in data:
            raise SchemaError("Missing 'eventID' in report schema.")

        if "clientName" not in data:
            raise SchemaError("Missing 'clientName' in report schema.")

        if "resources" not in data:
            raise SchemaError("Missing 'resources' in report schema.")

        self._id = data.get("id", None)
        self._created = None
        self._modified = None
        self._program_id = data["programID"]
        self._event_id = data["eventID"]
        self._client_name = data["clientName"]
        self._report_name = data.get("reportName", None)
        self._payload_descriptors = None
        self._resources = [ReportData(resource) for resource in data["resources"]]

        if "createdDateTime" in data:
            self._created = datetime.datetime.fromisoformat(data["createdDateTime"])

        if "modificationDateTime" in data:
            self._modified = datetime.datetime.fromisoformat(data["modificationDateTime"])

        if "payloadDescriptors" in data:
            self._payload_descriptors = [
                PayloadDescriptor(payload) for payload in data["payloadDescriptors"]
            ]

    @property
    def id(self) -> str | None:
        """The ID of the report."""
        return self._id

    @property
    def created(self) -> datetime.datetime | None:
        """The creation date of the report."""
        return self._created

    @property
    def modified(self) -> datetime.datetime | None:
        """The modification date of the report."""
        return self._modified

    @property
    def program_id(self) -> str:
        """The program ID of the report."""
        return self._program_id

    @property
    def event_id(self) -> str:
        """The event ID of the report."""
        return self._event_id

    @property
    def client_name(self) -> str:
        """The client name of the report."""
        return self._client_name

    @property
    def report_name(self) -> str | None:
        """The name of the report."""
        return self._report_name

    @property
    def payload_descriptors(self) -> list[PayloadDescriptor] | None:
        """The payload descriptors of the report."""
        return self._payload_descriptors

    @property
    def resources(self) -> list[ReportData]:
        """The resources of the report."""
        return self._resources



from .access_token import AccessToken, acquire_access_token
from .event import Event
from .events import TargetType, get_events
from .exceptions import SchemaError, ToadrError
from .interval import Interval
from .interval_period import IntervalPeriod
from .iso_date import parse_iso8601_duration
from .payload_descriptor import PayloadDescriptor
from .report import Report, ReportData
from .report_descriptor import ReportDescriptor
from .values_map import parse_values_map

__all__ = [
    "AccessToken",
    "acquire_access_token",
    "Event",
    "TargetType",
    "get_events",
    "SchemaError",
    "ToadrError",
    "Interval",
    "IntervalPeriod",
    "parse_iso8601_duration",
    "PayloadDescriptor",
    "Report",
    "ReportData",
    "ReportDescriptor",
    "parse_values_map",
]



import datetime
import os

import aiohttp

from .exceptions import ToadrError


class AccessToken:
    """Access token object.

    Enables tracking of expiration time and checking if the token is expired.
    """

    def __init__(self, token: str, expires_in: int):
        """Initialize the access token with an expiration time in seconds.

        Checking if the token is expired can be done with the is_expired() method. The
        is_expired() method will return True if the token is expired or will expire within
        the next 60 seconds.

        Parameters
        ----------
        token : str
            The access token.
        expires_in : int
            The time in seconds until the token expires.
        """
        self._token = token
        time_delta = datetime.timedelta(seconds=expires_in)
        self._expires_at = datetime.datetime.now(tz=datetime.timezone.utc) + time_delta

    @property
    def token(self) -> str:
        """The access token."""
        return self._token

    @property
    def expires_at(self) -> datetime:
        """The time in seconds until the token expires."""
        return self._expires_at

    @property
    def expires_in(self) -> int:
        """The time in seconds until the token expires."""
        expires_in = self._expires_at - datetime.datetime.now(tz=datetime.timezone.utc)
        return int(expires_in.total_seconds())

    def is_expired(self) -> bool:
        """Check if the token is expired."""
        time_delta = self._expires_at - datetime.datetime.now(tz=datetime.timezone.utc)
        return time_delta < datetime.timedelta(seconds=60)

    def __str__(self) -> str:
        """Return a string representation of the access token."""
        return f"{self.token}"

    def __repr__(self) -> str:
        """Return a string representation of the access token."""
        return f"AccessToken(token='{self.token}', expires_in={self.expires_in})"


async def acquire_access_token(
    session: aiohttp.ClientSession,
    token_url: str,
    grant_type: str,
    scope: str,
    client_id: str | None = None,
    client_secret: str | None = None,
) -> AccessToken:
    """Acquire an access token from the token provider.

    Connect to the token provider and acquire an access token. The access token will be returned
    as an AccessToken object. `client_id` and `client_secret` can be None if they are available
    in the environment as CLIENT_ID and CLIENT_SECRET.

    Parameters
    ----------
    session : aiohttp.ClientSession
        The aiohttp session to use for the request.
    token_url : str
        The URL to acquire the token from.
    grant_type : str
        The grant type to use for the token request.
    scope : str
        The scope of the token.
    client_id : str | None
        The client ID or None if acquirable from environment as CLIENT_ID.
    client_secret : str | None
        The client secret or None if acquirable from environment as CLIENT_SECRET.

    Returns
    -------
    AccessToken
        The access token object.

    Raises
    ------
    ValueError
        If the `client_id` or `client_secret` are not provided and not available in the environment.
    toadr3.ToadrError
        If the request to the token provider fails.
    """
    if client_id is None:
        if "CLIENT_ID" not in os.environ:
            raise ValueError("client_id is required")
        client_id = os.getenv("CLIENT_ID")

    if client_secret is None:
        if "CLIENT_SECRET" not in os.environ:
            raise ValueError("client_secret is required")
        client_secret = os.getenv("CLIENT_SECRET")

    if grant_type is None:
        raise ValueError("grant_type is required")

    if scope is None:
        raise ValueError("scope is required")

    credentials = {
        "grant_type": grant_type,
        "client_id": client_id,
        "client_secret": client_secret,
        "scope": scope,
    }

    async with session.post(token_url, data=credentials) as response:
        # 400 is start of HTTP error codes
        if response.status >= 400:  # noqa PLR2004 - Magic value used in comparison
            raise ToadrError(
                "Failed to acquire access token",
                status_code=response.status,
                reason=response.reason,
                headers=response.headers,
                json_response=await response.json(),
            )

        data = await response.json()
        return AccessToken(data["access_token"], data["expires_in"])



from typing import Any

from .exceptions import SchemaError
from .values_map import parse_values_map


class ReportDescriptor:
    """An object that may be used to request a report from a VEN.

    See OpenADR REST User Guide for detailed description of how configure a report request.
    """

    def __init__(self, data: dict):
        if "payloadType" not in data:
            raise SchemaError("Missing 'payloadType' in report descriptor schema.")

        self._payload_type: str = data["payloadType"]
        self._reading_type: str | None = data.get("readingType", None)
        self._units: str | None = data.get("units", None)
        self._targets: dict[str, Any] | None = None
        self._aggregate: bool = data.get("aggregate", False)
        self._start_interval: int = int(data.get("startInterval", -1))
        self._num_intervals: int = int(data.get("numIntervals", -1))
        self._historical: bool = data.get("historical", True)
        self._frequency: int = int(data.get("frequency", -1))
        self._repeat: int = int(data.get("repeat", 1))

        if "targets" in data:
            self._targets = parse_values_map(data["targets"])

    @property
    def payload_type(self) -> str:
        """The type of the payload."""
        return self._payload_type

    @property
    def reading_type(self) -> str | None:
        """The type of reading."""
        return self._reading_type

    @property
    def units(self) -> str | None:
        """The measurement units of the reading."""
        return self._units

    @property
    def targets(self) -> dict[str, Any] | None:
        """The targets of the report."""
        return self._targets

    @property
    def aggregate(self) -> bool:
        """Whether the report should be aggregated."""
        return self._aggregate

    @property
    def start_interval(self) -> int:
        """The interval at which to generate the report (-1 means end of last)."""
        return self._start_interval

    @property
    def num_intervals(self) -> int:
        """The number of intervals to include in the report."""
        return self._num_intervals

    @property
    def historical(self) -> bool:
        """Indicator for report data direction.

        True reports on intervals preceding startInterval,
        False reports on intervals following startInterval.
        """
        return self._historical

    @property
    def frequency(self) -> int:
        """Number of intervals between reports.

        -1 means the same as numIntervals.
        """
        return self._frequency

    @property
    def repeat(self) -> int:
        """The number of times to repeat the report.

        1 indicate generate one report, -1 indicate repeat indefinitely.
        """
        return self._repeat



from typing import Any

from .exceptions import SchemaError
from .interval_period import IntervalPeriod
from .values_map import parse_values_map


class Interval:
    """Interval object.

    Has a payload for an IntervalPeriod. The IntervalPeriod is either local to the class or
    available from the parent Event object.
    """

    def __init__(self, data: dict):
        """Create an interval object from parsed JSON data."""
        self._data = data

        if "id" not in data:
            raise SchemaError("Missing 'id' in interval schema.")

        if "payloads" not in data:
            raise SchemaError("Missing 'payloads' in interval schema.")

        self._id = data["id"]
        self._interval_period = None
        self._payloads = parse_values_map(data["payloads"])

        if "intervalPeriod" in data:
            self._interval_period = IntervalPeriod(data["intervalPeriod"])

    @property
    def id(self) -> int:
        """The ID of the interval."""
        return self._data["id"]

    def has_interval_period(self) -> bool:
        """Check if the interval has an interval period."""
        return self._interval_period is not None

    @property
    def interval_period(self) -> IntervalPeriod | None:
        """The interval period of the interval."""
        return self._interval_period

    @property
    def payloads(self) -> dict[str, Any]:
        """The payloads of the interval."""
        return self._payloads



import datetime

from .exceptions import SchemaError
from .iso_date import parse_iso8601_duration


class IntervalPeriod:
    """Interval period object.

    Represents a period of time with a start time, duration, and randomize start time.
    """

    def __init__(self, data: dict[str, str]):
        """Create an interval period object from parsed JSON data."""
        if "start" not in data:
            raise SchemaError("Missing 'start' in interval period schema.")

        self._start = datetime.datetime.fromisoformat(data["start"])
        self._duration = datetime.timedelta(seconds=0)
        self._randomize_start = datetime.timedelta(seconds=0)

        if "duration" in data:
            self._duration = parse_iso8601_duration(data["duration"])

        if "randomizeStart" in data:
            self._randomize_start = parse_iso8601_duration(data["randomizeStart"])

    @property
    def start(self) -> datetime.datetime:
        """The start of the interval period."""
        return self._start

    @property
    def duration(self) -> datetime.timedelta:
        """The duration of the interval period."""
        return self._duration

    @property
    def randomize_start(self) -> datetime.timedelta:
        """The randomize start of the interval period."""
        return self._randomize_start

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return (
            f"IntervalPeriod("
            f"start={self.start}, "
            f"duration={self.duration}, "
            f"randomize_start={self.randomize_start}"
            f")"
        )

    def __str__(self) -> str:
        """Return a description of the interval period."""
        return f"{self.start} - {self.start + self.duration} (± {self.randomize_start})"
